/* rdfdump.c - RDF Parser Toolkit (repat) example application
 *
 * Copyright (C) 2000 Jason Diamond - http://injektilo.org/
 *
 *    This package is Free Software available under either of two licenses:
 * 
 * 1. The GNU Lesser General Public License (LGPL)
 * 
 *    See http://www.gnu.org/copyleft/lesser.html or COPYING.LIB for the
 *    full license text.
 *      _________________________________________________________________
 * 
 *      Copyright (C) 2000 Jason Diamond. All Rights Reserved.
 * 
 *      This library is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU Lesser General Public License
 *      as published by the Free Software Foundation; either version 2 of
 *      the License, or (at your option) any later version.
 * 
 *      This library is distributed in the hope that it will be useful, but
 *      WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *      Lesser General Public License for more details.
 * 
 *      You should have received a copy of the GNU Lesser General Public
 *      License along with this library; if not, write to the Free Software
 *      Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *      USA
 *      _________________________________________________________________
 * 
 *    NOTE - under Term 3 of the LGPL, you may choose to license the entire
 *    library under the GPL. See COPYING for the full license text.
 * 
 * 2. The Mozilla Public License
 * 
 *    See http://www.mozilla.org/MPL/MPL-1.1.html or MPL.html for the full
 *    license text.
 * 
 *    Under MPL section 13. I declare that all of the Covered Code is
 *    Multiple Licensed:
 *      _________________________________________________________________
 * 
 *      The contents of this file are subject to the Mozilla Public License
 *      version 1.1 (the "License"); you may not use this file except in
 *      compliance with the License. You may obtain a copy of the License
 *      at http://www.mozilla.org/MPL/
 * 
 *      Software distributed under the License is distributed on an "AS IS"
 *      basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 *      the License for the specific language governing rights and
 *      limitations under the License.
 * 
 *      The Initial Developer of the Original Code is Jason Diamond.
 *      Portions created by Jason Diamond are Copyright (C) 2000 Jason
 *      Diamond. All Rights Reserved.
 * 
 *      Alternatively, the contents of this file may be used under the
 *      terms of the GNU Lesser General Public License, in which case the
 *      provisions of the LGPL License are applicable instead of those
 *      above. If you wish to allow use of your version of this file only
 *      under the terms of the LGPL License and not to allow others to use
 *      your version of this file under the MPL, indicate your decision by
 *      deleting the provisions above and replace them with the notice and
 *      other provisions required by the LGPL License. If you do not delete
 *      the provisions above, a recipient may use your version of this file
 *      under either the MPL or the LGPL License.
 */

/* this sample program reads an RDF document from stdin and
   dumps the statements and any containing or embedded XML
   to stdout. */

#include "rdfparse.h"

#include <stdio.h>
#include <string.h>
#include <limits.h>

#ifdef WIN32
#ifdef _DEBUG
#include <crtdbg.h>
#endif
#endif

/* prototypes */

	void
my_statement_handler(
	void*           user_data,
	RDF_SubjectType subject_type,
	const XML_Char* subject,
	const XML_Char* predicate,
	int             ordinal,
	RDF_ObjectType  object_type,
	const XML_Char* object,
	const XML_Char* xml_lang );

	void 
my_start_parse_type_literal_handler(
	void* user_data );

	void 
my_end_parse_type_literal_handler(
	void* user_data );

	void 
my_start_element_handler( 
	void* user_data, 
	const XML_Char* name, 
	const XML_Char** attributes );

	void
my_end_element_handler( 
	void* user_data, 
	const XML_Char* name );

	void 
my_character_data_handler( 
	void* user_data, 
	const XML_Char* s, 
	int len );

	void 
my_warning_handler(
	void* user_data,
	const XML_Char* warning );

/* main */

	int 
main(
	 int argc,
	 char** argv )
{
	FILE* input;
	char base[ PATH_MAX ];

	char buf[ BUFSIZ ];
	int done;

	int statements = 0;

	RDF_Parser rdf_parser;

#ifdef WIN32
#ifdef _DEBUG
	_CrtSetDbgFlag( _CrtSetDbgFlag( 0 ) | _CRTDBG_LEAK_CHECK_DF );
#endif
#endif

	if( argc > 1 )
	{
		input = fopen( argv[ 1 ], "r" );
		strcpy( base, "file:" );
		strcat( base, argv[ 1 ] );
	}
	else
	{
		input = stdin;
		strcpy( base, "stdin" );
	}

	rdf_parser = RDF_ParserCreate( NULL );

	RDF_SetUserData( rdf_parser, ( void* )&statements );

	RDF_SetStatementHandler( 
		rdf_parser, 
		my_statement_handler );

	RDF_SetParseTypeLiteralHandler( 
		rdf_parser, 
		my_start_parse_type_literal_handler,
		my_end_parse_type_literal_handler );

	RDF_SetElementHandler( 
		rdf_parser, 
		my_start_element_handler, 
		my_end_element_handler );

	RDF_SetCharacterDataHandler(
		rdf_parser,
		my_character_data_handler );

	RDF_SetWarningHandler( 
		rdf_parser, 
		my_warning_handler );

	RDF_SetBase( rdf_parser, base );

	do
	{
		size_t len = fread( buf, 1, sizeof( buf ), input );
		done = len < sizeof( buf );

		if ( ! RDF_Parse( rdf_parser, buf, len, done ) )
		{
			fprintf( 
				stderr,
				"**** ERROR **** : %s at line %d\n",
				XML_ErrorString( XML_GetErrorCode( RDF_GetXmlParser( rdf_parser ) ) ),
				XML_GetCurrentLineNumber( RDF_GetXmlParser( rdf_parser ) ) );

			return 1;
		}
	} 
	while( ! done );

	/* close file. */

	if( input != stdin )
	{
		fclose( input );
	}

	RDF_ParserFree( rdf_parser );

	printf( "Total statements: %d\n", statements );

	return 0;
}

/* handlers */

	void 
my_statement_handler(
	void*           user_data,
	RDF_SubjectType subject_type,
	const XML_Char* subject,
	const XML_Char* predicate,
	int             ordinal,
	RDF_ObjectType  object_type,
	const XML_Char* object,
	const XML_Char* xml_lang )
{
	int* statements = ( int* )user_data;

	++( *statements );

	printf( "triple(" );

	switch( subject_type )
	{
	case RDF_SUBJECT_TYPE_URI:
		printf( "\"%s\"", subject );
		break;
	case RDF_SUBJECT_TYPE_DISTRIBUTED:
		printf( "distributed(\"%s\")", subject );
		break;
	case RDF_SUBJECT_TYPE_PREFIX:
		printf( "prefix(\"%s\")", subject );
		break;
	case RDF_SUBJECT_TYPE_ANONYMOUS:
		printf( "\"%s\"", subject );
		break;
	}

	printf( ", \"%s\", ", predicate );

	switch( object_type )
	{
	case RDF_OBJECT_TYPE_RESOURCE:
		printf( "\"%s\"", object );
		break;
	case RDF_OBJECT_TYPE_LITERAL:
		printf( "literal(\"%s\")", object );
		break;
	case RDF_OBJECT_TYPE_XML:
		printf( "XML" );
		break;
	}

	printf( ")\n" );

}

	void 
my_start_parse_type_literal_handler(
	void* user_data )
{
	printf( "start parse type literal" );
}

	void 
my_end_parse_type_literal_handler(
	void* user_data )
{
	printf( "end parse type literal" );
}

	void 
my_start_element_handler( 
	void* user_data, 
	const XML_Char* name, 
	const XML_Char** attributes )
{
	printf( "start element: %s\n", name );
}

	void
my_end_element_handler( 
	void* user_data, 
	const XML_Char* name )
{
	printf( "end element: %s\n", name );
}

	void 
my_character_data_handler( 
	void* user_data, 
	const XML_Char* s, 
	int len )
{
	int i;

	printf( "characters: " );

	for( i = 0; i < len; ++i )
	{
		switch( s[ i ] )
		{
		case 0x09:
			printf( "&#x9;" );
			break;
		case 0x0A:
			printf( "&#xA;" );
			break;
		case 0x0D:
			printf( "&#xD;" );
			break;
		default:
			printf( "%c", s[ i ] );
		}
	}

	printf( "\n" );
}

	void 
my_warning_handler(
	void* user_data,
	const XML_Char* warning )
{
	printf( "**** WARNING **** : %s", warning );
}
