#include "rdfparse.h"

#include <stdio.h>
#include <string.h>
#include <limits.h>

#ifdef WIN32
#ifdef _DEBUG
#include <crtdbg.h>
#endif
#endif

/* prototypes */

	void
test_resolve_uri();

	void
assert_resolve_uri(
	const  RDF_Parser  parser,
	const  XML_Char*   reference_uri,
	const  XML_Char*   absolute_uri );

	void
test_parser(
	void**  data );

	void
test_statement_handler(
	void*           user_data,
	RDF_SubjectType subject_type,
	const XML_Char* subject,
	const XML_Char* predicate,
	int             ordinal,
	RDF_ObjectType  object_type,
	const XML_Char* object,
	const XML_Char* xml_lang );

	void 
test_start_parse_type_literal_handler(
	void* user_data );

	void 
test_end_parse_type_literal_handler(
	void* user_data );

	void 
test_start_element_handler( 
	void* user_data, 
	const XML_Char* name, 
	const XML_Char** attributes );

	void
test_end_element_handler( 
	void* user_data, 
	const XML_Char* name );

	void 
test_character_data_handler( 
	void* user_data, 
	const XML_Char* s, 
	int len );

	void 
test_warning_handler(
	void* user_data,
	const XML_Char* warning );

	void
test_string(
	const XML_Char*  test );

	void
test_int(
	int  test );

	void
test_subject_type(
	RDF_SubjectType  test );

	void
test_object_type(
	RDF_ObjectType  test );

/* test macros and data */

#define TEST_BEGIN( id, filename ) void* id[] = { filename,

#define TEST_STATEMENT(subject_type, subject, predicate, ordinal, object_type, object, xml_lang ) "statement", ( void* )subject_type, subject, predicate, ( void* )ordinal, ( void* )object_type, object, xml_lang,

#define TEST_WARNING "warning",


#define TEST_END "end" };

#include "rdftests.i"

/* globals */

void** current_data;
int current_index;
int current_failures;

/* main */
	int
main(
	 int argc,
	 char** argv )
{
#ifdef WIN32
#ifdef _DEBUG
	_CrtSetDbgFlag( _CrtSetDbgFlag( 0 ) | _CRTDBG_LEAK_CHECK_DF );
#endif
#endif

	test_resolve_uri();

	test_parser( about_01 );
	test_parser( about_02 );
	test_parser( about_03 );
	test_parser( about_04 );
	test_parser( ID_01 );
	test_parser( ID_02 );
	test_parser( ID_03 );
	test_parser( ID_04 );
	test_parser( container_01 );
	test_parser( container_02 );
	test_parser( container_03 );
	test_parser( container_04 );
	test_parser( attributes_01 );
	test_parser( attributes_02 );
	test_parser( attributes_03 );
	test_parser( statement_ID_01 );
	test_parser( anonymous_01 );
	test_parser( anonymous_02 );
	test_parser( anonymous_03 );
	test_parser( anonymous_04 );
	test_parser( anonymous_05 );
	test_parser( anonymous_06 );
	test_parser( resource_01 );
	test_parser( resource_02 );
	test_parser( resource_03 );
	test_parser( resource_04 );
	test_parser( bagID_01 );
}

	void
test_resolve_uri()
{
	RDF_Parser parser;
	XML_Char* base_uri = "http://a/b/c/d;p?q";

	parser = RDF_ParserCreate( NULL );
	
	RDF_SetBase( parser, base_uri );

	assert_resolve_uri( parser, "g:h",           "g:h"                   );
	assert_resolve_uri( parser, "g",             "http://a/b/c/g"        );
	assert_resolve_uri( parser, "./g",           "http://a/b/c/g"        );
	assert_resolve_uri( parser, "g/",            "http://a/b/c/g/"       );
	assert_resolve_uri( parser, "/g",            "http://a/g"            );
	assert_resolve_uri( parser, "//g",           "http://g"              );
	assert_resolve_uri( parser, "?y",            "http://a/b/c/?y"       );
	assert_resolve_uri( parser, "g?y",           "http://a/b/c/g?y"      );
	assert_resolve_uri( parser, "#s",            "http://a/b/c/d;p?q#s"  );
	assert_resolve_uri( parser, "g#s",           "http://a/b/c/g#s"      );
	assert_resolve_uri( parser, "g?y#s",         "http://a/b/c/g?y#s"    );
	assert_resolve_uri( parser, ";x",            "http://a/b/c/;x"       );
	assert_resolve_uri( parser, "g;x",           "http://a/b/c/g;x"      );
	assert_resolve_uri( parser, "g;x?y#s",       "http://a/b/c/g;x?y#s"  );
	assert_resolve_uri( parser, ".",             "http://a/b/c/"         );
	assert_resolve_uri( parser, "./",            "http://a/b/c/"         );
	assert_resolve_uri( parser, "..",            "http://a/b/"           );
	assert_resolve_uri( parser, "../",           "http://a/b/"           );
	assert_resolve_uri( parser, "../g",          "http://a/b/g"          );
	assert_resolve_uri( parser, "../..",         "http://a/"             );
	assert_resolve_uri( parser, "../../",        "http://a/"             );
	assert_resolve_uri( parser, "../../g",       "http://a/g"            );

	assert_resolve_uri( parser, "",              "http://a/b/c/d;p?q"    );

	assert_resolve_uri( parser, "../../../g",    "http://a/../g"         );
	assert_resolve_uri( parser, "../../../../g", "http://a/../../g"      );

	assert_resolve_uri( parser, "/./g",          "http://a/./g"          );
	assert_resolve_uri( parser, "/../g",         "http://a/../g"         );
	assert_resolve_uri( parser, "g.",            "http://a/b/c/g."       );
	assert_resolve_uri( parser, ".g",            "http://a/b/c/.g"       );
	assert_resolve_uri( parser, "g..",           "http://a/b/c/g.."      );
	assert_resolve_uri( parser, "..g",           "http://a/b/c/..g"      );

	assert_resolve_uri( parser, "./../g",        "http://a/b/g"          );
	assert_resolve_uri( parser, "./g/.",         "http://a/b/c/g/"       );
	assert_resolve_uri( parser, "g/./h",         "http://a/b/c/g/h"      );
	assert_resolve_uri( parser, "g/../h",        "http://a/b/c/h"        );
	assert_resolve_uri( parser, "g;x=1/./y",     "http://a/b/c/g;x=1/y"  );
	assert_resolve_uri( parser, "g;x=1/../y",    "http://a/b/c/y"        );

	assert_resolve_uri( parser, "g?y/./x",       "http://a/b/c/g?y/./x"  );
	assert_resolve_uri( parser, "g?y/../x",      "http://a/b/c/g?y/../x" );
	assert_resolve_uri( parser, "g#s/./x",       "http://a/b/c/g#s/./x"  );
	assert_resolve_uri( parser, "g#s/../x",      "http://a/b/c/g#s/../x" );

	assert_resolve_uri( parser, "http:g",        "http:g"                );

	assert_resolve_uri( parser, "g/../../../h",  "http://a/h"            );
}

	void
assert_resolve_uri(
	const  RDF_Parser  parser,
	const  XML_Char*   reference_uri,
	const  XML_Char*   absolute_uri )
{
	XML_Char buffer[ 256 ];

	RDF_ResolveURI(
		parser,
		reference_uri,
		buffer,
		sizeof( buffer ) );

	if( strcmp( buffer, absolute_uri ) != 0 )
	{
		printf( "FAIL: %s -> %s != %s\n", reference_uri, buffer, absolute_uri );
	}
}

	void
test_parser(
	void**  data )
{
	FILE* input;
	char base[ PATH_MAX ];

	char buf[ BUFSIZ ];
	int done;

	RDF_Parser rdf_parser;

	printf( "parsing %s\n", ( char* )data[ 0 ] );

	input = fopen( ( char* )data[ 0 ], "r" );
	strcpy( base, "file:" );
	strcat( base, ( char* )data[ 0 ] );

	rdf_parser = RDF_ParserCreate( NULL );

	current_data = data;
	current_index = 1;
	current_failures = 0;

	RDF_SetStatementHandler( 
		rdf_parser, 
		test_statement_handler );

	RDF_SetParseTypeLiteralHandler( 
		rdf_parser, 
		test_start_parse_type_literal_handler,
		test_end_parse_type_literal_handler );

	RDF_SetElementHandler( 
		rdf_parser, 
		test_start_element_handler, 
		test_end_element_handler );

	RDF_SetCharacterDataHandler(
		rdf_parser,
		test_character_data_handler );

	RDF_SetWarningHandler( 
		rdf_parser, 
		test_warning_handler );

	RDF_SetBase( rdf_parser, base );

	do
	{
		size_t len = fread( buf, 1, sizeof( buf ), input );
		done = len < sizeof( buf );

		if ( ! RDF_Parse( rdf_parser, buf, len, done ) )
		{
			fprintf( 
				stderr,
				"**** ERROR **** : %s at line %d\n",
				XML_ErrorString( XML_GetErrorCode( RDF_GetXmlParser( rdf_parser ) ) ),
				XML_GetCurrentLineNumber( RDF_GetXmlParser( rdf_parser ) ) );

			break;
		}
	} 
	while( ! done );

	test_string( "end" );

	printf( "  %d failures\n", current_failures );

	fclose( input );

	RDF_ParserFree( rdf_parser );
}

/* handlers */

	void 
test_statement_handler(
	void*           user_data,
	RDF_SubjectType subject_type,
	const XML_Char* subject,
	const XML_Char* predicate,
	int             ordinal,
	RDF_ObjectType  object_type,
	const XML_Char* object,
	const XML_Char* xml_lang )
{
	test_string( "statement" );
	test_subject_type( subject_type );
	test_string( subject );
	test_string( predicate );
	test_int( ordinal );
	test_object_type( object_type );
	test_string( object );
	test_string( xml_lang );
}

	void 
test_start_parse_type_literal_handler(
	void* user_data )
{
}

	void 
test_end_parse_type_literal_handler(
	void* user_data )
{
}

	void 
test_start_element_handler( 
	void* user_data, 
	const XML_Char* name, 
	const XML_Char** attributes )
{
}

	void
test_end_element_handler( 
	void* user_data, 
	const XML_Char* name )
{
}

	void 
test_character_data_handler( 
	void* user_data, 
	const XML_Char* s, 
	int len )
{
}

	void 
test_warning_handler(
	void* user_data,
	const XML_Char* warning )
{
	test_string( "warning" );
}

	void
test_string(
	const  XML_Char*  test )
{
	char* current;
	int result = 0;

	current = current_data[ current_index++ ];

	if( current == NULL || test == NULL )
	{
		if( current == NULL && test == NULL )
		{
			result = 0;
		}
		else
		{
			result = 1;
		}
	}
	else
	{
		result = strcmp( current, test );
	}

	if(  result != 0 )
	{
		++current_failures;

		printf( "    FAIL: %s != %s\n", current, test );
	}
}

	void
test_int(
	int  test )
{
	int current = ( int )current_data[ current_index++ ];

	if( current != test )
	{
		++current_failures;

		printf( "    FAIL: %d != %d\n", current, test );
	}
}

	const char*
get_subject_type_string(
	RDF_SubjectType subject_type )
{
	switch( subject_type )
	{
	case RDF_SUBJECT_TYPE_URI:
		return "RDF_SUBJECT_TYPE_URI";
	case RDF_SUBJECT_TYPE_DISTRIBUTED:
		return "RDF_SUBJECT_TYPE_DISTRIBUTED";
	case RDF_SUBJECT_TYPE_PREFIX:
		return "RDF_SUBJECT_TYPE_PREFIX";
	case RDF_SUBJECT_TYPE_ANONYMOUS:
		return "RDF_SUBJECT_TYPE_ANONYMOUS";
	}

	return "RDF_SUBJECT_TYPE_UNKNOWN";
}

	void
test_subject_type(
	RDF_SubjectType  test )
{
	RDF_SubjectType current = ( RDF_SubjectType )current_data[ current_index++ ];

	if( current != test )
	{
		++current_failures;

		printf( 
			"    FAIL: %s != %s\n", 
			get_subject_type_string( current ), 
			get_subject_type_string( test ) );
	}
}

	const char*
get_object_type_string(
	RDF_ObjectType object_type )
{
	switch( object_type )
	{
	case RDF_OBJECT_TYPE_RESOURCE:
		return "RDF_OBJECT_TYPE_RESOURCE";
	case RDF_OBJECT_TYPE_LITERAL:
		return "RDF_OBJECT_TYPE_LITERAL";
	case RDF_OBJECT_TYPE_XML:
		return "RDF_OBJECT_TYPE_XML";
	}

	return "RDF_OBJECT_TYPE_UNKNOWN";
}

	void
test_object_type(
	RDF_ObjectType  test )
{
	RDF_ObjectType current = ( RDF_ObjectType )current_data[ current_index++ ];

	if( current != test )
	{
		++current_failures;

		printf( 
			"    FAIL: %s != %s\n", 
			get_object_type_string( current ), 
			get_object_type_string( test ) );
	}
}
